#' @title Plot Procrustes variation
#'
#' @description Modification of the \code{\link[geomorph]{plotRefToTarget}} function to allow to display landmark variation on Procrustes
#'
#' @param M1 Matrix of landmark coordinates for the first (reference) specimen
#' @param M2 Matrix of landmark coordinates for the second (target) specimen
#' @param col Either a single color value (\code{"character"}), vector of values or function for colouring both points and vectors; or a list of two of any of these three elements. The first argument is passed to the points and the second to the vectors (\code{default = list("grey", "black")} for grey points and black vectors - see details).
#' @param col.val Optional, if \code{col} is a function for assigning colours (or a list of two functions), which values to pass to the function to assign the colors).
#' @param magnitude The magnitude of difference between the two shapes (\code{default = 1}).
#' @param labels Logical, whether to plot the landmark labels (\code{TRUE}) or not (\code{FALSE} - default).
#' @param axes  Logical, whether to plot the plot axes (\code{TRUE}) or not (\code{FALSE} - default).
#' @param pt.size The size of the points (\code{default = 1}).
#' @param gridPar Optional, a named list of grid parameters generated by \code{\link[geomorph]{gridPar}}.
#' @param ... Additional parameters to \code{\link[graphics]{plot}} or \code{\link[rgl]{plot3d}}.
#' 
#' @details
#' \code{col} arguments are passed to \code{\link[graphics]{plot}} or \code{\link[rgl]{plot3d}} as follows:
#' \itemize{
#'    \item if col is a single value or a single vector, it passes it to both points and colors arguments;
#'    \item if col is a single function, it applies it to both points and colors arguments based on the argument \code{col.val};
#'    \item if col is a list of two values, vectors or functions, it will apply the first element to the points and the second to the vectors the same way as described above;
#' }
#' 
#' 
#' @examples
#' ## Loading the geomorph dataset
#' require(geomorph)
#' data(plethodon)
#' 
#' ## Performing the Procrustes superimposition
#' proc_super <- gpagen(plethodon$land, print.progress = FALSE)
#' 
#' ## Getting the range of variation
#' variation <- variation.range(proc_super, return.ID = TRUE)
#' 
#' ## Selecting the coordinates and the variation vector
#' M1 <- proc_super$coords[, , variation$min.max[1]]
#' M2 <- proc_super$coords[, , variation$min.max[2]]
#' var_val <- variation$range[, 1]
#'  
#' ## Plot the variation
#' procrustes.var.plot(M1, M2)
#' 
#' ## A colourful plot
#' procrustes.var.plot(M1, M2, col = list(grDevices::rainbow, "pink"), col.val = var_val,
#'                     pt.size = 2.5)
#' 
#' \dontrun{
#' ## Loading the scallops 3D data from geomorph
#' require(geomorph)
#' data(scallops)
#' 
#' ## Procrustes superimposition
#' procrustes <- gpagen(scallops$coorddata)
#' 
#' ## Getting the range of variation
#' variation <- variation.range(procrustes, return.ID = TRUE)
#' 
#' ## Selecting the coordinates and the variation vector
#' M1 <- procrustes$coords[, , variation$min.max[1]]
#' M2 <- procrustes$coords[, , variation$min.max[2]]
#' var_val <- variation$range[, 1]
#'  
#' ## Plot the variation in 3D
#' procrustes.var.plot(M1, M2, col.val = var_val, col = list(grDevices::heat.colors, "grey"))
#' }
#' 
#' @seealso \code{\link{variation.range}}
#'  
#' @author Thomas Guillerme - original function by Dean Adams, Emma Sherratt & Michael Collyer.
#' 
#' @export 
#' @importFrom graphics plot segments points text arrows
#' @importFrom rgl plot3d text3d
#' @importFrom geomorph gridPar
#' @importFrom stats na.omit

procrustes.var.plot <- function(M1, M2, col = list("grey", "black"), col.val, magnitude = 1, labels = FALSE, axes = FALSE, pt.size = 1, gridPar, ...) {

  ## Check the inputs
  check.class(M1, "matrix")
  check.class(M2, "matrix")

  ## Get the input dimensions
  dim <- unique(c(ncol(M1), ncol(M2)))
  if(length(dim) != 1) {
    stop("M1 and M2 do not have the same number of dimensions!", .call = FALSE)
  }
  if(dim < 2 || dim > 3) {
    stop("M1 and M2 must be 2D or 3D matrices.", .call = FALSE)
  }

  ## Check M1 and M2 for missing data
  if(any(is.na(M1)) == TRUE) {
    stop("Data contains missing values. Estimate these first (see 'geomorph::estimate.missing').", .call = FALSE)
  }
  if(any(is.na(M2)) == TRUE) {
    stop("Data contains missing values. Estimate these first (see 'geomorph::estimate.missing').", .call = FALSE)
  }

  ## Check the colour arguments
  ## Is it in the right format?
  if(class(col) == "list") {
    col_classes <- unlist(lapply(col, class))
      if(!any(col_classes %in% c("character", "function"))) {
        stop("col argument must be a single value, vector (of the same number of rows as M1) or function or a list of two of any of the former.", .call = FALSE)
      }
    col_list <- TRUE
    ## Should have two elements
    if(length(col) > 2) {
      stop("col argument must be a single value, vector (of the same number of rows as M1) or function or a list of two of any of the former.", .call = FALSE)
    }
  } else {
    if(!any(class(col) %in% c("character", "function"))) {
      stop("col argument must be a single value, vector (of the same number of rows as M1) or function or a list of two of any of the former.", .call = FALSE)
    }
    col_list <- FALSE
    ## Converting into a list for hanging
    col <- list(col)
  }

  ## Checking the arguments
  col_fun_check <- rep(FALSE, length(col))
  for(sub in 1:length(col)) {
    ## Get the class
    class_col <- class(col[[sub]])
    ## Check the argument
    if(class_col != "function") {
      if(class_col == "character") {
        if(length(col[[sub]]) > 1) {
          if(nrow(M1) < length(col[[sub]])) {
            stop("col argument must be a single value, vector (of the same number of rows as M1) or function or a list of two of any of the former.", .call = FALSE)
          }
        } else {
          col[[sub]] <- rep(col[[sub]], nrow(M1))
        }
      } else {
        stop("col argument must be a single value, vector (of the same number of rows as M1) or function or a list of two of any of the former.", .call = FALSE)
      }
    } else {
      ## Colour is a function
      col_fun_check[sub] <- TRUE
    }
  }

  ## Checking the col.val argument
  if(!missing(col.val)) {
    if(class(col.val) == "list") {
      if(length(col.val) > 2) {
        stop("col.val should be a list of two vectors or a vector, all of the same length as the number of rows in M1 and M2.")
      }
    } else {
      col.val <- list(col.val, col.val)
    }

    ## Check the length
    col_val_length <- unlist(lapply(col.val, length))
    if(any(col_val_length != nrow(M1))) {
      stop("col.val should be a list of two vectors or a vector, all of the same length as the number of rows in M1 and M2.")
    } else {
      val.points <- col.val[[1]]
      val.vector <- col.val[[2]]
    }
  } else {
    if(any(col_fun_check)) {
      stop("col.val argument is missing with no default.")
    }
  }

  ## Check the magnitude
  check.class(magnitude, c("integer", "numeric"))
  check.length(magnitude, 1, " must be a single numeric value.")
  if(magnitude != 1) {
    ## Scale M2
    M2 <- M2 + (M2 - M1) * magnitude
  }

  ## Check the labels
  check.class(labels, "logical")

  ## Check the axes
  check.class(axes, "logical")

  ## Getting the grid parameters
  grid_par <- geomorph::gridPar()

  ## Adding any non-default gridPar
  if(!missing(gridPar)) {
    check.class(gridPar, "list")

    ## Matching the parameters names 
    default_names <- names(grid_par)
    overwrite_names <- names(gridPar)
    match_names <- match(default_names, overwrite_names)

    ## Overwriting the parameters
    if(!all(is.na(match_names))) {
      grid_par[which(!is.na(match_names))] <- gridPar[stats::na.omit(match_names)]
    }
  }

  ## Check the point sizes
  check.class(pt.size, c("integer", "numeric"))
  check.length(pt.size, 1, " must be a single numeric value.")
  grid_par$pt.size <- pt.size


  ## Function for getting the limit (from internal geomorph::plotRefToTarget)
  limits <- function(matrix, offset) { 
    mat_range <- range(matrix)
    scale_range <- scale(mat_range, scale = FALSE)
    return(mean(mat_range) + offset * scale_range)
  }

  ## Colour gradient function
  get.colour.gradient <- function(values, col.fun) {

    ## Sort the data by range
    histo <- hist(values, plot = FALSE)
    n_col <- length(histo$counts)

    ## Get the gradient
    avail_cols <- rev(col.fun(n_col))

    ## Attribute the colours
    col_out <- rep(avail_cols[n_col], length(values))

    ## Attribute the colour by range
    for(colour in rev(1:n_col)) {
      col_out[(values <= histo$breaks[colour])] <- avail_cols[colour]
    }

    return(col_out)
  }

  ## Handling the colour arguments
  if(col_list) {
    if(col_fun_check[1]) {
      ## Apply the function
      col_points <- get.colour.gradient(val.points, col[[1]])
    } else {
      col_points <- col[[1]]
    }
    if(col_fun_check[2]) {
      ## Apply the function
      col_vector <- get.colour.gradient(val.vector, col[[2]])
    } else {
      col_vector <- col[[2]]
    }
  } else {
    if(any(col_fun_check)) {
      col_points <- col_vector <- get.colour.gradient(val.points, col[[1]])
    } else {
      col_points <- col_vector <- col[[1]]
    }
  }

  ## 2D plots
  if(dim == 2) {

    if(axes) {
      graphics::plot(M1, asp = 1, type = "n", xlab = "x", ylab = "y", xlim = limits(M1[, 1], 1.25), ylim = limits(M1[, 2], 1.25), ...)
    } else {
      graphics::plot(M1, asp = 1, type = "n", xlab = "", ylab = "", xlim = limits(M1[, 1], 1.25), axes = FALSE, ylim = limits(M1[, 2], 1.25), ...)
    }

    if(labels) {
      graphics::text(M1, labels = paste(1:dim(M1)[1]), adj = grid_par$txt.adj, pos = grid_par$txt.pos, cex = grid_par$txt.cex, col = grid_par$txt.col)
    }

    graphics::arrows(M1[, 1], M1[, 2], M2[, 1], M2[, 2], length = 0.075, lwd = 2, col = col_vector)

    graphics::points(M1, pch = 21, bg = col_points, cex = grid_par$pt.size, col = grid_par$pt.bg)


  return(invisible())

}

  ## 3D plots
  if(dim == 3) {

    if(axes) {
      rgl::plot3d(M1, type = "s", col = col_points, size = grid_par$pt.size, aspect = FALSE, ...)
    } else {
      rgl::plot3d(M1, type = "s", col = col_points, size = grid_par$pt.size, aspect = FALSE, xlab = "", ylab = "", zlab = "", axes = FALSE, ...)
    }
    if(labels) {
      rgl::text3d(M1, texts = paste(1:dim(M1)[1]), adj = (grid_par$txt.adj + grid_par$pt.size), pos = (grid_par$txt.pos + grid_par$pt.size), cex = grid_par$txt.cex, col = grid_par$txt.col)
    } 

    ## Function for adding the segments
    add.segment <- function(i, M1, M2, lwd = 2, col) {
      rgl::segments3d(rbind(M1[i, ], M2[i, ]), lwd = lwd, col = col[i])
    }

    silent <- sapply(1:nrow(M1), add.segment, M1 = M1, M2 = M2, col = col_vector)

    # for (i in 1:nrow(M1)) {
    #   silent <- rgl::segments3d(rbind(M1[i, ], M2[i, ]), lwd = 2, col = col_vector[i])
    # }
  
    return(invisible())

  }
}
